@@ -17,20 +17,16 @@ from typing import (
     Awaitable,
     Callable,
     Coroutine,
-    Dict,
     Generator,
     Iterable,
     Iterator,
-    List,
-    Tuple,
-    Union,
     cast,
 )
 
 if sys.version_info >= (3, 10):
-    from typing import Concatenate, ParamSpec
+    from typing import Concatenate, ParamSpec, TypeAlias
 else:
-    from typing_extensions import Concatenate, ParamSpec
+    from typing_extensions import Concatenate, ParamSpec, TypeAlias
 
 from contextlib import asynccontextmanager
 
@@ -43,18 +39,15 @@ from .types import PayloadType, T
 MQTT_LOGGER = logging.getLogger("mqtt")
 MQTT_LOGGER.setLevel(logging.WARNING)
 
-_PahoSocket = Union[socket.socket, ssl.SSLSocket, mqtt.WebsocketWrapper, Any]
+_PahoSocket: TypeAlias = "socket.socket | ssl.SSLSocket | mqtt.WebsocketWrapper | Any"
 
-WebSocketHeaders = Union[
-    Dict[str, str],
-    Callable[[Dict[str, str]], Dict[str, str]],
-]
+WebSocketHeaders: TypeAlias = (
+    "dict[str, str] | Callable[[dict[str, str]], dict[str, str]]"
+)
 
 
 class ProtocolVersion(IntEnum):
-    """
-    A mapping of Paho MQTT protocol version constants to an Enum for use in type hints.
-    """
+    """A mapping of paho-mqtt protocol versions to an Enum for use in type hints."""
 
     V31 = mqtt.MQTTv31
     V311 = mqtt.MQTTv311
@@ -103,17 +96,12 @@ class ProxySettings:
 
 
 # See the overloads of `socket.setsockopt` for details.
-SocketOption = Union[
-    Tuple[int, int, Union[int, bytes]],
-    Tuple[int, int, None, int],
-]
+SocketOption: TypeAlias = "tuple[int, int, int | bytes] | tuple[int, int, None, int]"
 
-SubscribeTopic = Union[
-    str,
-    Tuple[str, mqtt.SubscribeOptions],
-    List[Tuple[str, mqtt.SubscribeOptions]],
-    List[Tuple[str, int]],
-]
+SubscribeTopic: TypeAlias = (
+    "str | tuple[str, mqtt.SubscribeOptions] | list[tuple[str, mqtt.SubscribeOptions]]"
+    " | list[tuple[str, int]]"
+)
 
 P = ParamSpec("P")
 
@@ -134,8 +122,111 @@ def _outgoing_call(
     return decorated
 
 
-class Client:
+@dataclass(frozen=True)
+class Wildcard:
+    """A topic, optionally with wildcards (+ and #). Can only be subscribed to."""
+
+    value: str
+
+    def __str__(self) -> str:
+        return self.value
+
+    def __post_init__(self) -> None:
+        """Validate the wildcard."""
+        if not isinstance(self.value, str):
+            raise TypeError("wildcard must be a string")
+        if (
+            len(self.value) == 0
+            or len(self.value) > 65535
+            or "#/" in self.value
+            or any(
+                "+" in level or "#" in level
+                for level in self.value.split("/")
+                if len(level) > 1
+            )
+        ):
+            raise ValueError(f"Invalid wildcard: {self.value}")
+
+
+WildcardLike: TypeAlias = "str | Wildcard"
+
+
+@dataclass(frozen=True)
+class Topic(Wildcard):
+    """A topic that can be published and subscribed to."""
+
+    def __post_init__(self) -> None:
+        """Validate the topic."""
+        if not isinstance(self.value, str):
+            raise TypeError("topic must be a string")
+        if (
+            len(self.value) == 0
+            or len(self.value) > 65535
+            or "+" in self.value
+            or "#" in self.value
+        ):
+            raise ValueError(f"Invalid topic: {self.value}")
+
+    def matches(self, wildcard: WildcardLike) -> bool:
+        """Check if the topic is matched by a given wildcard."""
+        if not isinstance(wildcard, Wildcard):
+            wildcard = Wildcard(wildcard)
+        # Split topics into levels to compare them one by one
+        topic_levels = self.value.split("/")
+        wildcard_levels = str(wildcard).split("/")
+        if wildcard_levels[0] == "$share":
+            # Shared subscriptions use the topic structure: $share/<group_id>/<topic>
+            wildcard_levels = wildcard_levels[2:]
+
+        def recurse(x: list[str], y: list[str]) -> bool:
+            if not x:
+                if not y:
+                    return True
+                return False
+            if not y:
+                return False
+            if y[0] == "#":
+                return True
+            if x[0] == y[0] or y[0] == "+":
+                return recurse(x[1:], y[1:])
+            return False
+
+        return recurse(topic_levels, wildcard_levels)
+
+
+TopicLike: TypeAlias = "str | Topic"
+
+
+class Message:
+    """Custom message class that allows us to use our own Topic class."""
+
     def __init__(
+        self,
+        topic: TopicLike,
+        payload: PayloadType,
+        qos: int,
+        retain: bool,
+        mid: int,
+    ):
+        self.topic = Topic(topic) if not isinstance(topic, Topic) else topic
+        self.payload = payload
+        self.qos = qos
+        self.retain = retain
+        self.mid = mid
+
+    @classmethod
+    def _from_paho_message(cls, message: mqtt.MQTTMessage) -> Message:
+        return cls(
+            topic=message.topic,
+            payload=message.payload,
+            qos=message.qos,
+            retain=message.retain,
+            mid=message.mid,
+        )
+
+
+class Client:
+    def __init__(  # noqa: C901
         self,
         hostname: str,
         port: int = 1883,
@@ -183,6 +274,12 @@ class Client:
         self._pending_calls_threshold: int = 10
         self._misc_task: asyncio.Task[None] | None = None
 
+        # List of all callbacks to call when a message is received
+        self._on_message_callbacks: list[Callable[[Message], None]] = []
+        self._unfiltered_messages_callback: Callable[
+            [mqtt.Client, Any, mqtt.MQTTMessage], None
+        ] | None = None
+
         self._outgoing_calls_sem: asyncio.Semaphore | None
         if max_concurrent_outgoing_calls is not None:
             self._outgoing_calls_sem = asyncio.Semaphore(max_concurrent_outgoing_calls)
@@ -203,7 +300,7 @@ class Client:
         self._client.on_disconnect = self._on_disconnect
         self._client.on_subscribe = self._on_subscribe
         self._client.on_unsubscribe = self._on_unsubscribe
-        self._client.on_message = None
+        self._client.on_message = self._on_message
         self._client.on_publish = self._on_publish
         # Callbacks for custom event loop
         self._client.on_socket_open = self._on_socket_open
@@ -249,7 +346,9 @@ class Client:
         self._socket_options = tuple(socket_options)
 
     @property
-    def id(self) -> str:
+    def id(  # noqa: A003 # TODO: When doing BREAKING CHANGES rename to avoid shadowing builtin id
+        self,
+    ) -> str:
         """Return the client ID.
 
         Note that paho-mqtt stores the client ID as `bytes` internally.
@@ -260,9 +359,7 @@ class Client:
 
     @property
     def _pending_calls(self) -> Generator[int, None, None]:
-        """
-        Yield all message IDs with pending calls.
-        """
+        """Yield all message IDs with pending calls."""
         yield from self._pending_subscribes.keys()
         yield from self._pending_unsubscribes.keys()
         yield from self._pending_publishes.keys()
@@ -290,7 +387,7 @@ class Client:
         # We convert all of them to the common MqttError for user convenience.
         # See: https://github.com/eclipse/paho.mqtt.python/blob/v1.5.0/src/paho/mqtt/client.py#L1770
         except (OSError, mqtt.WebsocketConnectionError) as error:
-            raise MqttError(str(error))
+            raise MqttError(str(error)) from None
         await self._wait_for(self._connected, timeout=timeout)
 
     async def disconnect(self, *, timeout: int = 10) -> None:
@@ -323,12 +420,12 @@ class Client:
         if result != mqtt.MQTT_ERR_SUCCESS:
             raise MqttCodeError(result, "Could not subscribe to topic")
         # Create future for when the on_subscribe callback is called
-        cb_result: asyncio.Future[
+        callback_result: asyncio.Future[
             tuple[int] | list[mqtt.ReasonCodes]
         ] = asyncio.Future()
-        with self._pending_call(mid, cb_result, self._pending_subscribes):
-            # Wait for cb_result
-            return await self._wait_for(cb_result, timeout=timeout)
+        with self._pending_call(mid, callback_result, self._pending_subscribes):
+            # Wait for callback_result
+            return await self._wait_for(callback_result, timeout=timeout)
 
     @_outgoing_call
     async def unsubscribe(
@@ -380,26 +477,20 @@ class Client:
     async def filtered_messages(
         self, topic_filter: str, *, queue_maxsize: int = 0
     ) -> AsyncGenerator[AsyncGenerator[mqtt.MQTTMessage, None], None]:
-        """Return async generator of messages that match the given filter.
-
-        Use queue_maxsize to restrict the queue size. If the queue is full,
-        incoming messages will be discarded (and a warning is logged).
-        If queue_maxsize is less than or equal to zero, the queue size is infinite.
-
-        Example use:
-            async with client.filtered_messages('floors/+/humidity') as messages:
-                async for message in messages:
-                    print(f'Humidity reading: {message.payload.decode()}')
-        """
-        cb, generator = self._cb_and_generator(
+        """Return async generator of messages that match the given filter."""
+        MQTT_LOGGER.warning(
+            "filtered_messages() is deprecated and will be removed in a future version."
+            " Use messages() together with Topic.matches() instead."
+        )
+        callback, generator = self._deprecated_callback_and_generator(
             log_context=f'topic_filter="{topic_filter}"', queue_maxsize=queue_maxsize
         )
         try:
-            self._client.message_callback_add(topic_filter, cb)
+            self._client.message_callback_add(topic_filter, callback)
             # Back to the caller (run whatever is inside the with statement)
             yield generator
         finally:
-            # We are exitting the with statement. Remove the topic filter.
+            # We are exiting the with statement. Remove the topic filter.
             self._client.message_callback_remove(topic_filter)
 
     @asynccontextmanager
@@ -407,24 +498,47 @@ class Client:
         self, *, queue_maxsize: int = 0
     ) -> AsyncGenerator[AsyncGenerator[mqtt.MQTTMessage, None], None]:
         """Return async generator of all messages that are not caught in filters."""
+        MQTT_LOGGER.warning(
+            "unfiltered_messages() is deprecated and will be removed in a future"
+            " version. Use messages() instead."
+        )
         # Early out
-        if self._client.on_message is not None:
-            # TODO: This restriction can easily be removed.
+        if self._unfiltered_messages_callback is not None:
             raise RuntimeError(
                 "Only a single unfiltered_messages generator can be used at a time."
             )
-        cb, generator = self._cb_and_generator(
+        callback, generator = self._deprecated_callback_and_generator(
             log_context="unfiltered", queue_maxsize=queue_maxsize
         )
         try:
-            self._client.on_message = cb
+            self._unfiltered_messages_callback = callback
             # Back to the caller (run whatever is inside the with statement)
             yield generator
         finally:
-            # We are exitting the with statement. Unset the callback.
-            self._client.on_message = None
+            # We are exiting the with statement. Unset the callback.
+            self._unfiltered_messages_callback = None
 
-    def _cb_and_generator(
+    @asynccontextmanager
+    async def messages(
+        self, *, queue_maxsize: int = 0
+    ) -> AsyncGenerator[AsyncGenerator[Message, None], None]:
+        """Return async generator of incoming messages.
+
+        Use queue_maxsize to restrict the queue size. If the queue is full,
+        incoming messages will be discarded (and a warning is logged).
+        If queue_maxsize is less than or equal to zero, the queue size is infinite.
+        """
+        callback, generator = self._callback_and_generator(queue_maxsize=queue_maxsize)
+        try:
+            # Add to the list of callbacks to call when a message is received
+            self._on_message_callbacks.append(callback)
+            # Back to the caller (run whatever is inside the with statement)
+            yield generator
+        finally:
+            # We are exiting the with statement. Remove the callback from the list.
+            self._on_message_callbacks.remove(callback)
+
+    def _deprecated_callback_and_generator(
         self, *, log_context: str, queue_maxsize: int = 0
     ) -> tuple[
         Callable[[mqtt.Client, Any, mqtt.MQTTMessage], None],
@@ -434,10 +548,10 @@ class Client:
         messages: asyncio.Queue[mqtt.MQTTMessage] = asyncio.Queue(maxsize=queue_maxsize)
         # Callback for the underlying API
         def _put_in_queue(
-            client: mqtt.Client, userdata: Any, msg: mqtt.MQTTMessage
+            client: mqtt.Client, userdata: Any, message: mqtt.MQTTMessage
         ) -> None:
             try:
-                messages.put_nowait(msg)
+                messages.put_nowait(message)
             except asyncio.QueueFull:
                 MQTT_LOGGER.warning(
                     f"[{log_context}] Message queue is full. Discarding message."
@@ -473,13 +587,53 @@ class Client:
 
         return _put_in_queue, _message_generator()
 
+    def _callback_and_generator(
+        self, *, queue_maxsize: int = 0
+    ) -> tuple[Callable[[Message], None], AsyncGenerator[Message, None]]:
+        # Queue to hold the incoming messages
+        messages: asyncio.Queue[Message] = asyncio.Queue(maxsize=queue_maxsize)
+
+        def _callback(message: Message) -> None:
+            """Put the new message in the queue."""
+            try:
+                messages.put_nowait(message)
+            except asyncio.QueueFull:
+                MQTT_LOGGER.warning("Message queue is full. Discarding message.")
+
+        async def _generator() -> AsyncGenerator[Message, None]:
+            """Forward all messages from the message queue."""
+            while True:
+                # Wait until we either:
+                #  1. Receive a message
+                #  2. Disconnect from the broker
+                get: asyncio.Task[Message] = self._loop.create_task(messages.get())
+                try:
+                    done, _ = await asyncio.wait(
+                        (get, self._disconnected), return_when=asyncio.FIRST_COMPLETED
+                    )
+                except asyncio.CancelledError:
+                    # If the asyncio.wait is cancelled, we must make sure
+                    # to also cancel the underlying tasks.
+                    get.cancel()
+                    raise
+                if get in done:
+                    # We received a message. Return the result.
+                    yield get.result()
+                else:
+                    # We got disconnected from the broker. Cancel the "get" task.
+                    get.cancel()
+                    # Stop the generator with the following exception
+                    raise MqttError("Disconnected during message iteration")
+
+        return _callback, _generator()
+
     async def _wait_for(
         self, fut: Awaitable[T], timeout: float | None, **kwargs: Any
     ) -> T:
         try:
             return await asyncio.wait_for(fut, timeout=timeout, **kwargs)
         except asyncio.TimeoutError:
-            raise MqttError("Operation timed out")
+            raise MqttError("Operation timed out") from None
 
     @contextmanager
     def _pending_call(
@@ -579,7 +733,7 @@ class Client:
         userdata: Any,
         mid: int,
         properties: mqtt.Properties | None = None,
-        reasonCodes: list[mqtt.ReasonCodes] | mqtt.ReasonCodes | None = None,
+        reason_codes: list[mqtt.ReasonCodes] | mqtt.ReasonCodes | None = None,
     ) -> None:
         try:
             self._pending_unsubscribes.pop(mid).set()
@@ -588,6 +742,17 @@ class Client:
                 f'Unexpected message ID "{mid}" in on_unsubscribe callback'
             )
 
+    def _on_message(
+        self, client: mqtt.Client, userdata: Any, message: mqtt.MQTTMessage
+    ) -> None:
+        # Call the deprecated unfiltered_messages callback
+        if self._unfiltered_messages_callback is not None:
+            self._unfiltered_messages_callback(client, userdata, message)
+        # Convert the paho.mqtt message into our own Message type
+        m = Message._from_paho_message(message)
+        for callback in self._on_message_callbacks:
+            callback(m)
+
     def _on_publish(self, client: mqtt.Client, userdata: Any, mid: int) -> None:
         try:
             self._pending_publishes.pop(mid).set()
@@ -600,7 +765,7 @@ class Client:
     def _on_socket_open(
         self, client: mqtt.Client, userdata: Any, sock: _PahoSocket
     ) -> None:
-        def cb() -> None:
+        def callback() -> None:
             # client.loop_read() may raise an exception, such as BadPipe. It's
             # usually a sign that the underlaying connection broke, therefore we
             # disconnect straight away
@@ -610,19 +775,18 @@ class Client:
                 if not self._disconnected.done():
                     self._disconnected.set_exception(exc)
 
-        self._loop.add_reader(sock.fileno(), cb)
+        self._loop.add_reader(sock.fileno(), callback)
         # paho-mqtt calls this function from the executor thread on which we've called
-        # `self._client.connect()` (see [3]), so we create a callback function to schedule
-        # `_misc_loop()` and run it on the loop thread-safely.
-        def create_task_cb() -> None:
+        # `self._client.connect()` (see [3]), so we create a callback function to
+        # schedule `_misc_loop()` and run it on the loop thread-safely.
+        def create_task_callback() -> None:
             self._misc_task = self._loop.create_task(self._misc_loop())
 
-        self._loop.call_soon_threadsafe(create_task_cb)
+        self._loop.call_soon_threadsafe(create_task_callback)
 
     def _on_socket_close(
         self, client: mqtt.Client, userdata: Any, sock: _PahoSocket
     ) -> None:
-
         fileno = sock.fileno()
         if fileno > -1:
             self._loop.remove_reader(fileno)
@@ -632,7 +796,7 @@ class Client:
     def _on_socket_register_write(
         self, client: mqtt.Client, userdata: Any, sock: _PahoSocket
     ) -> None:
-        def cb() -> None:
+        def callback() -> None:
             # client.loop_write() may raise an exception, such as BadPipe. It's
             # usually a sign that the underlaying connection broke, therefore we
             # disconnect straight away
@@ -642,7 +806,7 @@ class Client:
                 if not self._disconnected.done():
                     self._disconnected.set_exception(exc)
 
-        self._loop.add_writer(sock, cb)
+        self._loop.add_writer(sock, callback)
 
     def _on_socket_unregister_write(
         self, client: mqtt.Client, userdata: Any, sock: _PahoSocket
@@ -679,7 +843,8 @@ class Client:
         except MqttError as error:
             # We tried to be graceful. Now there is no mercy.
             MQTT_LOGGER.warning(
-                f'Could not gracefully disconnect due to "{error}". Forcing disconnection.'
+                f'Could not gracefully disconnect due to "{error}". Forcing'
+                " disconnection."
             )
             await self.force_disconnect()
 
